{
    MP3Player - Free Delphi Component
    Copyright (C) 1999  Kei Ishida

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit MP3Trds;

//  킯킩񂭂Ȃ̂ŁAXbh֌Wʃjbgɕ
//  pthread ӎĂ邩B

interface

uses
  Windows, Messages, SysUtils, Classes;

type
  TMPAbstractMutex = class(TObject)
  protected
  public
    procedure Lock; virtual; abstract;
    procedure Unlock; virtual; abstract;
  end;

  TMPSimpleMutex = class(TMPAbstractMutex)
  private
    FCS: TRTLCriticalSection;
  protected
  public
    constructor Create;
    destructor Destroy; override;

    procedure Lock;  override;
    procedure Unlock;  override;
  end;

  TMPCondVar = class(TObject)
  private
    FMutex: TMPAbstractMutex;
    FEvent: THandle;
  public
    constructor Create(Mutex: TMPAbstractMutex);
    destructor Destroy; override;

    procedure Lock;
    procedure Unlock;
    procedure Wait;
//    procedure Notify;
    procedure NotifyAll;
  end;

  TMPThread = class;

  TMPThreadProc = procedure(Thread: TMPThread) of object;

  TMPThread = class(TObject)
  private
    FHandle: THandle;
    FThreadID: LongWord;
    FProc: TMPThreadProc;

    procedure SetPriority(const Value: Integer);
    function GetPriority: Integer;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Run(Proc: TMPThreadProc);

    property Priority: Integer read GetPriority write SetPriority;
  end;

  TMPReadAheadStream = class(TStream)
  private
    FThread: TMPThread;

    FOriginal: TStream;

    FBuffer: PChar;
    FBufSize: Integer;
    FBufAvailableLen: Integer;

    FBufMutex: TMPSimpleMutex;
    FBufEaterCondVar: TMPCondVar;
    FBufFillerCondVar: TMPCondVar;
    FBufFilling: Boolean;
    FStopBufFill: Boolean;

    FReadAheadMutex: TMPSimpleMutex;

    procedure BufFillProc(AThread: TMPThread);
  protected
  public
    constructor Create(BufSize: Integer);
    destructor Destroy; override;

    procedure Open(Original: TStream);
    procedure Close;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

implementation

{ TMPAbstractMutex }

{ TMPSimpleMutex }

constructor TMPSimpleMutex.Create;
begin
  inherited Create;

  InitializeCriticalSection(FCS);
end;

destructor TMPSimpleMutex.Destroy;
begin
  DeleteCriticalSection(FCS);

  inherited Destroy;
end;

procedure TMPSimpleMutex.Lock;
begin
  EnterCriticalSection(FCS);
end;

procedure TMPSimpleMutex.Unlock;
begin
  LeaveCriticalSection(FCS);
end;


{ TMPCondVar }

constructor TMPCondVar.Create(Mutex: TMPAbstractMutex);
begin
  Assert(Mutex <> nil);

  inherited Create;

  FMutex := Mutex;

  FEvent := CreateEvent(nil, False, False, nil);
end;

destructor TMPCondVar.Destroy;
begin
  CloseHandle(FEvent);

  inherited Destroy;
end;

procedure TMPCondVar.Lock;
begin
  FMutex.Lock;
end;

procedure TMPCondVar.NotifyAll;
begin
  SetEvent(FEvent);
end;

procedure TMPCondVar.Unlock;
begin
  FMutex.Unlock;
end;

procedure TMPCondVar.Wait;
begin
  ResetEvent(FEvent);
  FMutex.Unlock;
  WaitForSingleObject(FEvent, INFINITE);
  FMutex.Lock;
end;

{ TMPThread }

function ThreadProc(Parent: TMPThread): Integer;
begin
  Result := 0;
  try
    Parent.FProc(Parent);
  finally
    EndThread(Result);
  end;
end;

constructor TMPThread.Create;
begin
  inherited Create;

  FHandle := BeginThread(nil, 0, @ThreadProc, Self, CREATE_SUSPENDED, FThreadId );
end;

destructor TMPThread.Destroy;
begin
  if FHandle <> 0 then begin
    WaitForSingleObject(FHandle, INFINITE);
    CloseHandle(FHandle);
  end;

  inherited Destroy;
end;

function TMPThread.GetPriority: Integer;
begin
  Result := GetThreadPriority(FHandle);
end;

procedure TMPThread.Run(Proc: TMPThreadProc);
begin
  FProc := Proc;
  ResumeThread(FHandle);
end;

procedure TMPThread.SetPriority(const Value: Integer);
begin
  SetThreadPriority(FHandle, Value);
end;

{ TMPReadAheadStream }

procedure TMPReadAheadStream.BufFillProc(AThread: TMPThread);
const
  defaultReadLen = 4*1024;
var
  readLen: Integer;
  buf: array [0..defaultReadLen-1] of Char;
begin
  while True do begin
    FReadAheadMutex.Lock;
    try
      try
        FBufMutex.Lock;
        try
          while not((FBufFilling and (FBufAvailableLen < FBufSize)) or
                     FStopBufFill) do begin
            FReadAheadMutex.Unlock;
            FBufFillerCondVar.Wait;
            FReadAheadMutex.Lock;
          end;

          if FStopBufFill then begin
            Exit;
          end;

          if defaultReadLen <= FBufSize-FBufAvailableLen then begin
            readLen := defaultReadLen;
          end
          else begin
            readLen := FBufSize-FBufAvailableLen;
          end;
        finally
          FBufMutex.Unlock;
        end;

        readLen := FOriginal.Read(buf, readLen);

        FBufMutex.Lock;
        try
          if 0 < readLen then begin
            Move(buf[0], FBuffer[FBufAvailableLen], readLen);
            Inc(FBufAvailableLen, readLen);
          end
          else begin
            FBufFilling := False;
          end;
        finally
          FBufMutex.Unlock;
          FBufEaterCondVar.NotifyAll;
        end;
      except
        on e: Exception do begin
          FBufFilling := False;
          ShowException(e, ExceptAddr);
          FBufEaterCondVar.NotifyAll;
        end;
      end;
    finally
      FReadAheadMutex.Unlock;
    end;
  end;
end;

procedure TMPReadAheadStream.Close;
begin
  FBufMutex.Lock;
  try
    FBufFilling := False;
  finally
    FBufMutex.Unlock;
  end;

  if Assigned(FBuffer) then begin
    FreeMem(FBuffer);
    FBuffer := nil;
  end;

  FOriginal := nil;
end;

constructor TMPReadAheadStream.Create(BufSize: Integer);
begin
  Assert(0 < BufSize);

  inherited Create;

  FBufSize := BufSize;

  FBufMutex := TMPSimpleMutex.Create;
  FBufFillerCondVar := TMPCondVar.Create(FBufMutex);
  FBufEaterCondVar := TMPCondVar.Create(FBufMutex);

  FReadAheadMutex := TMPSimpleMutex.Create;

  FThread := TMPThread.Create;
  FThread.Run(BufFillProc);
end;

destructor TMPReadAheadStream.Destroy;
begin
  Close;

  FBufMutex.Lock;
  try
    FStopBufFill := True;
    FBufFilling := False;
  finally
    FBufMutex.Unlock;
    FBufFillerCondVar.NotifyAll;
    FBufEaterCondVar.NotifyAll;
  end;

  FThread.Free;

  FBufFillerCondVar.Free;
  FBufEaterCondVar.Free;
  FBufMutex.Free;

  FReadAheadMutex.Free;

  inherited Destroy;
end;

procedure TMPReadAheadStream.Open(Original: TStream);
begin
  Assert(Original <> nil);

  Close;

  FOriginal := Original;

  GetMem(FBuffer, FBufSize);

  FBufAvailableLen := 0;

  FBufFillerCondVar.Lock;
  try
    FBufFilling := True;
  finally
    FBufFillerCondVar.Unlock;
    FBufFillerCondVar.NotifyAll;
  end;
end;

function TMPReadAheadStream.Read(var Buffer; Count: Integer): Longint;
var
  readLen: Integer;
begin
  Result := 0;

  while 0 < Count do begin
    FBufMutex.Lock;
    try
      while not((0 < FBufAvailableLen) or (not FBufFilling)) do begin
        FBufEaterCondVar.Wait;
      end;
      if 0 < FBufAvailableLen then begin
        if FBufAvailableLen < Count then begin
          readLen := FBufAvailableLen;
        end
        else begin
          readLen := Count;
        end;
        Move(FBuffer[0], PChar(@Buffer)[Result], readLen);
        Move(FBuffer[readLen], FBuffer[0], FBufAvailableLen-readLen);
        Dec(Count, readLen);
        Inc(Result, readLen);
        Dec(FBufAvailableLen, readLen);
      end
      else begin
        Exit;
      end;
    finally
      FBufMutex.Unlock;
      FBufFillerCondVar.NotifyAll;
    end;
  end;
end;

function TMPReadAheadStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  FReadAheadMutex.Lock;
  try
    case Origin of
      soFromCurrent: begin
        Result := FOriginal.Seek(Offset-FBufAvailableLen, Origin);
        if Offset <> 0 then begin
          FBufAvailableLen := 0;
        end;
      end;
    else
      Result := FOriginal.Seek(Offset, Origin);
      FBufAvailableLen := 0;
    end;
    FBufFilling := True;
  finally
    FReadAheadMutex.Unlock;
    FBufFillerCondVar.NotifyAll;
  end;
end;

function TMPReadAheadStream.Write(const Buffer; Count: Integer): Longint;
begin
  Result := 0;
end;

end.
